from setuptools import setup

setup(
	name='viseng',
	version='0.1.1',
	packages=['viseng'],
	url='https://github.com/OllieBoyne/viseng',
	license='',
	author='ollie',
	author_email='ollieboyne@gmail.com',
	description='Visualisation library for producing graphics from Python\'s Matplotlib'
)
